package gophersmash

import (
	"encoding/json"
	"io/ioutil"
)

// STRUCTURES
// -----------
// The following structures are intended to
// load raw Json data directly into memory.
//
// The struct maps almost directly into a
// MashSketch struct, except for the MinHash.
//
// The Json stores the signatures, but we have
// to create a Minhash object and populate its
// signatures manually.

// MashJsonCollection is a struct that is used to
// load Mash MinHash collections dumped by Mash.
type mashJsonCollection struct {

	// Parameters
	Params mashJsonParameters

	// "sketches" key maps to a list of nested JSONs,
	// so define json key maps within the struct
	Sketches []mashJsonSketch `json:"sketches"`
}

// A struct to store a set of parameters loaded
// from a Mash Json dump file.
type mashJsonParameters struct {
	KmerSize     int    `json:"kmer"`
	Alphabet     string `json:"alphabet"`
	Canonical    bool   `json:"canonical"`
	PreserveCase bool   `json:"preserveCase"`
	HashType     string `json:"hashType"`
	HashBits     int    `json:"hashBits"`
	HashSeed     uint32 `json:"hashSeed"`
	SketchSize   int    `json:"sketchSize"`
}

// A struct to store the list of sketches found in
// the Json file. These sketches are not useful by
// themselves, they require a mash parameters struct.
type mashJsonSketch struct {
	Name    string   `json:"name"`
	Length  int      `json:"length"`
	Comment string   `json:"comment"`
	Hashes  []uint64 `json:"hashes"`
}

// METHODS:
// -----------
//
// The following methods extend the functionality of
// the structs defined above.

// Convert a mashJsonSketch struct
// to a MinHashMetadata struct.
func (sketch *mashJsonSketch) AsMinHashMetadata() *MinHashMetadata {
	return &MinHashMetadata{
		Name:        sketch.Name,    // A brief name describing the MinHash
		Description: sketch.Comment, // A longer description of the MinHash
		Length:      sketch.Length,  // The cumulative length of all inputs to the MinHash
	}
}

// Convert a mashJsonparameters
// struct (specific to Mash data
// loaded from a Json) to a
// MinHashParameters struct (more
// universal).
func (mashp *mashJsonParameters) AsMinHashParameters() *MinHashParameters {
	return &MinHashParameters{
		KmerSize:     mashp.KmerSize,
		Alphabet:     mashp.Alphabet,
		Canonical:    mashp.Canonical,
		PreserveCase: mashp.PreserveCase,
		HashType:     mashp.HashType,
		HashBits:     mashp.HashBits,
		HashSeed:     mashp.HashSeed,
		Nhashes:      mashp.SketchSize,
	}
}

// FUNCTIONS
// -----------
//
// The following functions are related to
// loading and saving Mash MinHash sketches
// in Json format.

// Load a Json file containing a collection of Mash
// MinHashes, populate MashSketch structs from them,
// and return the MashSketch structs in a slice.
func LoadMashJson(filename string) ([]*MinHash, error) {

	// Read the file and get content
	file, _ := ioutil.ReadFile(filename)
	content := []byte(string(file))

	// Load parameters from the file
	collection := mashJsonCollection{}

	// Note that Unmarshaling magic is due to the
	// `json:""` directives above.

	// We need to unmarshal the Json into the
	// Params struct (within mashMinHash) first.
	json.Unmarshal(content, &collection.Params)

	// Now we can unmarshal the rest.
	json.Unmarshal(content, &collection)

	// Finished loading Json data into memory.
	// Now create MinHash objects for each sketch.
	all_sketches := make([]*MinHash, len(collection.Sketches))
	for i, sketch := range collection.Sketches {

		// Add Name/Length/Comment
		// to MinHashMetadata struct

		// Create a Minhash object from the seed and
		// signatures in the Json file
		mh, _ := NewMinHashFromSignatures(
			collection.Params.AsMinHashParameters(),
			sketch.AsMinHashMetadata(),
			sketch.Hashes,
		)
		all_sketches[i] = mh
	}
	return all_sketches, nil
}
