package gophersmash

import (
	"errors"
	"fmt"
	"math"

	set "github.com/emirpasic/gods/sets/hashset"
	"github.com/spaolacci/murmur3"
)

/*
gopher-smash MinHash:

This file contains structs, methods, and functions
for using MinHashes. This particular implementation
uses a k-minimum values approach, following
[1], [2], and [3].

[1] https://github.com/dgryski/go-minhash

[2] https://godoc.org/github.com/dgryski/go-minhash

[3] https://research.neustar.biz/2012/07/09/sketch-of-the-day-k-minimum-values/
*/

// STRUCTURES:
//
// Note: the rpc package in Go's stdlib provides a
// template for using structs to store sets of
// parameters, and get around the inability to
// use keyword arguments.

// 64-bit hash function takes input string and seed,
// and returns a 64-bit unsigned integer.
type Hash64 func(string, uint32) uint64

// MinHash parameters that are common to an entire
// collection of genomic MinHashes.
type MinHashParameters struct {
	// Genomic minhash params
	KmerSize     int    // The size of kmers used to generate the signatures
	Alphabet     string // The alphabet that composes the kmers
	Canonical    bool   // Whether to use the canonical kmer (use whichever comes lexicographically first, kmer or kmer reverse complement)
	PreserveCase bool   // Whether to preserve the case of kmers being added
	HashType     string // Type of hash function to use
	HashBits     int    // Number of bits of the resulting hashes
	HashSeed     uint32 // Seed for the hash function
	Nhashes      int    // Number of signatures to keep/use
}

// Struct to store metadata for a MinHash.
// Keep this fungible, please.
type MinHashMetadata struct {
	Name        string // Name for this set of DNA
	Description string // Description of this set of DNA
	Length      int    // Total length of genome/DNA
	// md5???
}

// MinHash structure used to store a genomic MinHash.
// Use NewMinHash() to make a new one.
type MinHash struct {
	Params       *MinHashParameters // Pointer to parameters for the collection
	KmerSize     int                // The size of kmers used to generate the signatures
	HashFunction Hash64             // The hash function
	HashSeed     uint32             // Seed for the hash function
	Hashes       []uint64           // List of hashes that are part of this MinHash
	MaxHash      uint64             // Maximum hash value in the MinHeap
}

// METHODS:

// Validate MinHashParameters struct
func (p *MinHashParameters) Validate() error {
	if p.KmerSize < 1 {
		msg := "Error: MinHash parameter kmer size must be a positive nonzero number."
		return errors.New(msg)
	}
	if p.HashBits != 64 {
		msg := "Error: MinHash parameter hash bits must be 64"
		return errors.New(msg)
	}
	if p.HashType != "MurmurHash3_64_128" {
		msg := "Error: MinHash parameter hash type must be 'MurmurHash3_64_128'"
		return errors.New(msg)
	}
	return nil
}

// Check if two MinHashParameters are equal
func (p *MinHashParameters) Equals(q *MinHashParameters) bool {
	if p.KmerSize != q.KmerSize {
		return false
	}
	if p.Alphabet != q.Alphabet {
		return false
	}
	if p.Canonical != q.Canonical {
		return false
	}
	if p.PreserveCase != q.PreserveCase {
		return false
	}
	if p.HashType != q.HashType {
		return false
	}
	if p.HashBits != q.HashBits {
		return false
	}
	if p.HashSeed != q.HashSeed {
		return false
	}
	if p.Nhashes != q.Nhashes {
		return false
	}
	return true
}

// Hash a string and push it into the data structure
// using the k-minimum value approach.
// https://research.neustar.biz/2012/07/09/sketch-of-the-day-k-minimum-values/
func (m *MinHash) Push(s string) {
	h := m.HashFunction(s, m.HashSeed)
	for i, v := range m.Hashes {
		if h < v {
			m.Hashes[i] = h
			if m.MaxHash < h {
				m.MaxHash = h
			}
			break
		}
	}
}

// Check if this MinHash is equal to another MinHash
// (includes parameters and signatures, not metadata).
func (m *MinHash) Equals(n *MinHash) bool {

	// Check if these MinHashes were initialized
	// with the same MinHash parameters.
	if !m.Params.Equals(n.Params) {
		return false
	}

	// Check that the signatures (set of hashes)
	// are equal.
	if !m.SignatureEquals(n) {
		return false
	}

	return true
}

// Check if this MinHash's signature (set of hashed inputs)
// is equal to another MinHash's signature (does not include
// parameters or metadata).
func (m *MinHash) SignatureEquals(n *MinHash) bool {

	// Lengths must be equal
	if len(m.Hashes) != len(n.Hashes) {
		return false
	}

	// Create a set with each of the hashes
	mHashesSet := set.New()
	nHashesSet := set.New()
	for i := 0; i < len(m.Hashes); i++ {
		mHashesSet.Add(m.Hashes[i])
		nHashesSet.Add(n.Hashes[i])
	}

	// Number of unique elements must be equal
	if mHashesSet.Size() != nHashesSet.Size() {
		return false
	}

	// Something something,
	// if A subset B
	// and |A| == |B|
	// implies B subset A
	for _, k := range mHashesSet.Values() {
		if !nHashesSet.Contains(k) {
			return false
		}
	}

	return true
}

// Get the signature (list of hashes) from the given
// MinHash.
func (m *MinHash) Signature() []uint64 {
	return m.Hashes
}

// FUNCTIONS:

// Compute the hash of a string using MurmurHash3.
// This computes the 128-bit MurmurHash3 and only
// keeps/returns the first 64 bits.
func MurmurHash3_x64_128(s string, seed uint32) uint64 {
	h128 := murmur3.New128WithSeed(seed)
	h128.Write([]byte(s))
	v1, _ := h128.Sum128()
	return v1
}

// This takes a MinHash Parameters struct (loaded from
// Json or created by the user) as an input, and
// creates and returns a new MinHash struct.
func NewMinHash(params *MinHashParameters, metadata *MinHashMetadata) (*MinHash, error) {
	return NewMinHashFromSignatures(params, metadata, []uint64{})
}

// This takes a MinHash Parameters struct (loaded from
// Json or created by the user) as an input, and
// creates and returns a new MinHash struct.
func NewMinHashFromSignatures(params *MinHashParameters, metadata *MinHashMetadata, signatures []uint64) (*MinHash, error) {

	// Validate MinHash params
	err := params.Validate()
	if err != nil {
		return nil, err
	}

	// Validate MinHash metadata
	if metadata.Name == "" {
		msg := fmt.Sprintf("Error: MinHash metadata was missing a name, or name is an empty string.")
		return nil, errors.New(msg)
	}

	// Size check
	if len(signatures) > params.Nhashes {
		msg := fmt.Sprintf("Error: New MinHash function was passed too many signatures: Number of MinHash signatures = %d, Length of signatures passed = %d",
			params.Nhashes, len(signatures))
		return nil, errors.New(msg)
	}

	// Initialize the list of minhash signatures
	minimums := make([]uint64, params.Nhashes)

	// Set the max hash value
	maxHash := uint64(0)

	// Copy signatures into minimums
	for i := 0; i < params.Nhashes; i++ {
		if i < len(signatures) {
			// Update max hash
			if maxHash < signatures[i] {
				maxHash = signatures[i]
			}
			// Copy signature
			minimums[i] = signatures[i]
		} else {
			// New MinHash has extra slots for signatures.
			// sourmash does not pre-fill this array,
			// but we fill it with max uint64.
			minimums[i] = math.MaxUint64
		}
	}

	// This map stores a list of hash functions that our MinHash
	// implementation can utilize under the hood.
	hashFuncMap := map[string]Hash64{
		"MurmurHash3_64_128": MurmurHash3_x64_128,
	}

	// Validate() already made sure the user only
	// specified defined hash functions in the
	// MinHash parameters.
	hashFunc := hashFuncMap[params.HashType]

	return &MinHash{
		Params:       params,
		KmerSize:     params.KmerSize,
		HashFunction: hashFunc,
		HashSeed:     params.HashSeed,
		Hashes:       minimums,
		MaxHash:      maxHash,
	}, nil
}
