package gophersmash

import (
	"fmt"
	"testing"
)

// Test that MurmurHash of "AAA" returns the same thing as sourmash.
func TestMurmurHash(t *testing.T) {
	s := "AAA"
	seed := uint32(42)
	hash := MurmurHash3_x64_128(s, seed)
	gold := uint64(10679328328772601858)
	if hash != gold {
		msg := fmt.Sprintf("Error: Computing MurmurmurHash for %s: MurmurHash returned wrong value.", s)
		t.Error(msg)
	}
}

// Test initialization of a MinHashParameters struct and
// initialization of a new MinHash struct from it.
func TestInit(t *testing.T) {
	params := &MinHashParameters{
		KmerSize:     21,
		Alphabet:     "ATGC",
		Canonical:    true,
		PreserveCase: false,
		HashType:     "MurmurHash3_64_128",
		HashBits:     64,
		HashSeed:     42,
		Nhashes:      1,
	}
	metadata := &MinHashMetadata{
		Name:        "Test Name",
		Description: "Test Description",
	}
	_, err := NewMinHash(params, metadata)
	if err != nil {
		t.Error(err)
	}
}

// Test initialization of a MinHashParameters struct and
// initialization of a new MinHash struct from it
// and from a slice of signatures.
func TestInitFromSigs(t *testing.T) {
	params := &MinHashParameters{
		KmerSize:     21,
		Alphabet:     "ATGC",
		Canonical:    true,
		PreserveCase: false,
		HashType:     "MurmurHash3_64_128",
		HashBits:     64,
		HashSeed:     42,
		Nhashes:      1,
	}
	metadata := &MinHashMetadata{
		Name:        "Test Name",
		Description: "Test Description",
	}
	signatures := []uint64{10679328328772601858}
	_, err := NewMinHashFromSignatures(params, metadata, signatures)
	if err != nil {
		t.Error(err)
	}
}

// Test ability to compare MinHashParameters structs.
func TestParamEquivalency(t *testing.T) {

	// Create two MinHash objects with matching signatures
	// but different metadata, and ensure they are still
	// found equal.
	//
	// Each MinHash requires a set of parameters,
	// and metadata describing the MinHash.

	// MinHash p
	p := &MinHashParameters{
		KmerSize:     21,
		Alphabet:     "ATGC",
		Canonical:    true,
		PreserveCase: false,
		HashType:     "MurmurHash3_64_128",
		HashBits:     64,
		HashSeed:     42,
		Nhashes:      1,
	}

	// MinHash q
	q := &MinHashParameters{
		Nhashes:      1,
		HashSeed:     42,
		HashBits:     64,
		HashType:     "MurmurHash3_64_128",
		PreserveCase: false,
		Canonical:    true,
		Alphabet:     "ATGC",
		KmerSize:     21,
	}

	// Test equivalency
	if !p.Equals(q) {
		t.Error("Error: checking Equals() with two equivalent MinHashParameters returned incorrect result")
	}
}

// Test ability to add new kmers to MinHashes and
// ability to compare MinHashes.
func TestInitEquivalency(t *testing.T) {

	p := &MinHashParameters{
		KmerSize:     21,
		Alphabet:     "ATGC",
		Canonical:    true,
		PreserveCase: false,
		HashType:     "MurmurHash3_64_128",
		HashBits:     64,
		HashSeed:     42,
		Nhashes:      1,
	}
	pmeta := &MinHashMetadata{
		Name:        "Test Name",
		Description: "Test Description",
	}

	signatures := []uint64{10679328328772601858} // this is the murmurhash of "AAA"
	mh1, err := NewMinHashFromSignatures(p, pmeta, signatures)
	if err != nil {
		t.Error(err)
	}

	mh2, err := NewMinHash(p, pmeta)
	mh2.Push("AAA") // this has a hash of 10679...
	if err != nil {
		t.Error(err)
	}

	// Make sure their signatures alone match
	if !mh1.SignatureEquals(mh2) {
		t.Error("Error: checking SignatureEquals() with two equivalent MinHashes returned incorrect result")
	}

	// Make sure their parameters and signatures match
	if !mh1.Equals(mh2) {
		t.Error("Error: checking Equals() with two equivalent MinHashes returned incorrect result")
	}
}
